local Portrait = {}
ShadowUF:RegisterModule(Portrait, "portrait", ShadowUFLocals["Portrait"])

-- If the camera isn't reset OnShow, it'll show the entire character instead of just the head, odd I know
local function resetCamera(self)
	if( UnitIsVisible(self.parent.unit) and UnitIsConnected(self.parent.unit) ) then
		self:SetCamera(0)
	end
end

local function resetGUID(self)
	self.guid = nil
end

-- Keeps portraits looking fine and dandy even if someone is out of range/offline and comes back
local function portraitMonitor(self, elapsed)
	self.timeElapsed = self.timeElapsed + elapsed
	if( self.timeElapsed <= 10 ) then return end
	self.timeElapsed = self.timeElapsed - 10
	
	Portrait:Update(self.parent)
end

function Portrait:OnEnable(frame)
	frame:RegisterUnitEvent("UNIT_PORTRAIT_UPDATE", self, "Update")
	frame:RegisterUpdateFunc(self, "UpdateFunc")
end

function Portrait:OnDisable(frame)
	frame:UnregisterAll(self)
end

function Portrait:OnPreLayoutApply(frame)
	if( frame.visibility.portrait ) then
		if( ShadowUF.db.profile.units[frame.unitType].portrait.type == "3D" ) then
			if( not frame.portraitModel ) then
				frame.portraitModel = CreateFrame("PlayerModel", nil, frame)
				frame.portraitModel:SetScript("OnShow", resetCamera)
				frame.portraitModel:SetScript("OnHide", resetGUID)
				frame.portraitModel.parent = frame
				
				-- No sense in doing an update for the player as they is never out of range of themselve
				if( frame.unit ~= "player" ) then
					frame.portraitModel:SetScript("OnUpdate", portraitMonitor)
					frame.portraitModel.timeElapsed = 0
				end
			end
			
			frame.portrait = frame.portraitModel
			frame.portrait.guid = nil
			frame.portrait:Show()

			ShadowUF.Layout:ToggleVisibility(frame.portraitTexture, false)
		else
			frame.portraitTexture = frame.portraitTexture or frame:CreateTexture(nil, "ARTWORK")
			frame.portrait = frame.portraitTexture
			frame.portrait:Show()

			ShadowUF.Layout:ToggleVisibility(frame.portraitModel, false)
		end
	end
end

function Portrait:UpdateFunc(frame)
	-- Portrait models can't be updated unless the GUID changed or else you have the animation jumping around
	if( ShadowUF.db.profile.units[frame.unitType].portrait.type == "3D" ) then
		local guid = UnitGUID(frame.unitOwner)
		if( frame.portrait.guid ~= guid ) then
			self:Update(frame)
		end
		
		frame.portrait.guid = guid
	else
		self:Update(frame)
	end
end

function Portrait:Update(frame)
	-- Use class thingy
	if( ShadowUF.db.profile.units[frame.unitType].portrait.type == "class" ) then
		local classToken = select(2, UnitClass(frame.unitOwner))
		if( classToken ) then
			frame.portrait:SetTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes")
			frame.portrait:SetTexCoord(CLASS_ICON_TCOORDS[classToken][1], CLASS_ICON_TCOORDS[classToken][2], CLASS_ICON_TCOORDS[classToken][3], CLASS_ICON_TCOORDS[classToken][4])
		else
			frame.portrait:SetTexture("")
		end
	-- Use 2D character image
	elseif( ShadowUF.db.profile.units[frame.unitType].portrait.type == "2D" ) then
		frame.portrait:SetTexCoord(0.10, 0.90, 0.10, 0.90)
		SetPortraitTexture(frame.portrait, frame.unitOwner)
	-- Use animated 3D portrait
	elseif( UnitIsVisible(frame.unitOwner) and UnitIsConnected(frame.unitOwner) ) then
		frame.portrait:SetUnit(frame.unitOwner)
		frame.portrait:SetCamera(0)
	-- Use 3D question mark
	else
		frame.portrait:SetModelScale(4.25)
		frame.portrait:SetPosition(0, 0, -1.5)
		frame.portrait:SetModel("Interface\\Buttons\\talktomequestionmark.mdx")	
	end
end




